@@ -193,6 +193,21 @@ class Thread
     Kernel.raise ThreadError, "Thread#priority= primitive failed"
   end
 
+  def name
+    Truffle.primitive :thread_get_name
+    Kernel.raise ThreadError, "Thread#name primitive failed"
+  end
+
+  def name=(val)
+    unless val.nil?
+      val = Rubinius::Type.check_null_safe(StringValue(val))
+      raise ArgumentError, "ASCII incompatible encoding #{val.encoding.name}" unless val.encoding.ascii_compatible?
+      # TODO BJF Aug 27, 2016 Need to rb_str_new_frozen the val here and SET_ANOTHER_THREAD_NAME
+    end
+    Truffle.invoke_primitive :thread_set_name, self, val
+    val
+  end
+
   def inspect
     stat = status()
     stat = "dead" unless stat
@@ -352,15 +367,38 @@ class Thread
 end
 
 class ThreadGroup
-
-  attr_reader :list
-
   def initialize
-    @list = []
+    @enclosed = false
+  end
+
+  def enclose
+    @enclosed = true
+  end
+
+  def enclosed?
+    @enclosed
   end
 
   def add(thread)
-    @list.push thread
+    raise ThreadError, "can't move to the frozen thread group" if self.frozen?
+    raise ThreadError, "can't move to the enclosed thread group" if self.enclosed?
+
+    from_tg = thread.group
+    return nil unless from_tg
+    raise ThreadError, "can't move from the frozen thread group" if from_tg.frozen?
+    raise ThreadError, "can't move from the enclosed thread group" if from_tg.enclosed?
+
+    Truffle.invoke_primitive :thread_set_group, thread, self
+    self
   end
 
+  def list
+    Thread.list.select { |th| th.group == self }
+  end
+
+  Default = ThreadGroup.new
+
 end
+Truffle.invoke_primitive :thread_set_group, Thread.current, ThreadGroup::Default
+
+
