@@ -15,7 +15,8 @@ module Truffle
 
       Object.class_eval do
         define_method(name.to_sym) do |*args|
-          Truffle::Interop.execute(method, *args)
+          ret = Truffle::Interop.execute(method, *args)
+          Truffle::Interop.from_java_string(ret)
         end
       end
     end
@@ -24,6 +25,18 @@ module Truffle
       export(name.to_s, Object.method(name.to_sym))
     end
 
+    def self.ruby_object_keys(object)
+      if object.is_a?(Hash)
+        object.keys.map do |key|
+          Truffle::Interop.to_java_string(key)
+        end
+      else
+        object.instance_variables.map do |key|
+          Truffle::Interop.to_java_string(key[1..-1])
+        end
+      end
+    end
+
     class ForeignEnumerable
       include Enumerable
 
@@ -49,6 +62,10 @@ module Truffle
       ForeignEnumerable.new(foreign)
     end
 
+    class Foreign
+
+    end
+
   end
 
 end
