@@ -32,6 +32,7 @@ extern "C" {
 // Support
 
 #define RUBY_CEXT (void *)truffle_import_cached("ruby_cext")
+#define MUST_INLINE __attribute__((always_inline))
 
 // Configuration
 
@@ -52,11 +53,7 @@ extern "C" {
 
 // Basic types
 
-#ifdef JT_INT_VALUE
-  typedef uintptr_t VALUE;
-#else
-  typedef void *VALUE;
-#endif
+typedef void *VALUE;
 
 typedef VALUE ID;
 
@@ -64,18 +61,31 @@ typedef VALUE ID;
 
 NORETURN(VALUE rb_f_notimplement(int args_count, const VALUE *args, VALUE object));
 
+// Non-standard
+
+NORETURN(void rb_jt_error(const char *message));
+
+void *rb_jt_to_native_handle(VALUE managed);
+VALUE rb_jt_from_native_handle(void *native);
+
 // Memory
 
-#define xmalloc     malloc
-#define xfree       free
-#define ruby_xfree  free
+#define xmalloc       malloc
+#define xfree         free
+#define ruby_xfree    free
+#define ruby_xcalloc  calloc
 
 #define ALLOC_N(type, n)            ((type *)malloc(sizeof(type) * (n)))
 #define ALLOCA_N(type, n)           ((type *)alloca(sizeof(type) * (n)))
 
-void *rb_alloc_tmp_buffer(volatile VALUE *buffer_pointer, long length);
-void *rb_alloc_tmp_buffer2(volatile VALUE *buffer_pointer, long count, size_t size);
-void rb_free_tmp_buffer(volatile VALUE *buffer_pointer);
+#define RB_ZALLOC_N(type, n)        ((type *)ruby_xcalloc((n), sizeof(type)))
+#define RB_ZALLOC(type)             (RB_ZALLOC_N(type, 1))
+#define ZALLOC_N(type, n)           RB_ZALLOC_N(type, n)
+#define ZALLOC(type)                RB_ZALLOC(type)
+
+void *rb_alloc_tmp_buffer(VALUE *buffer_pointer, long length);
+void *rb_alloc_tmp_buffer2(VALUE *buffer_pointer, long count, size_t size);
+void rb_free_tmp_buffer(VALUE *buffer_pointer);
 
 #define RB_ALLOCV(v, n)             rb_alloc_tmp_buffer(&(v), (n))
 #define RB_ALLOCV_N(type, v, n)     rb_alloc_tmp_buffer2(&(v), (n), sizeof(type))
@@ -349,11 +359,36 @@ VALUE rb_str_new_cstr(const char *string);
 VALUE rb_str_cat(VALUE string, const char *to_concat, long length);
 VALUE rb_str_cat2(VALUE string, const char *to_concat);
 VALUE rb_str_to_str(VALUE string);
-VALUE rb_string_value(volatile VALUE *value_pointer);
+
+MUST_INLINE VALUE rb_string_value(VALUE *value_pointer) {
+  VALUE value = *value_pointer;
+
+  if (!RB_TYPE_P(value, T_STRING)) {
+    value = rb_str_to_str(value);
+    *value_pointer = value;
+  }
+
+  return value;
+}
+
+MUST_INLINE char *rb_string_value_ptr(VALUE *value_pointer) {
+  VALUE string = rb_string_value(value_pointer);
+  return RSTRING_PTR(string);
+}
+
+MUST_INLINE char *rb_string_value_cstr(VALUE *value_pointer) {
+  VALUE string = rb_string_value(value_pointer);
+
+  if (!truffle_invoke_b(RUBY_CEXT, "rb_string_value_cstr_check", string)) {
+    rb_jt_error("rb_string_value_cstr failure case not implemented");
+    abort();
+  }
+
+  return RSTRING_PTR(string);
+}
+
 #define StringValue(value) rb_string_value(&(value))
 #define SafeStringValue StringValue
-char *rb_string_value_ptr(volatile VALUE* value_pointer);
-char *rb_string_value_cstr(volatile VALUE* value_pointer);
 #define StringValuePtr(string) rb_string_value_ptr(&(string))
 #define StringValueCStr(string) rb_string_value_cstr(&(string))
 VALUE rb_str_buf_new(long capacity);
@@ -395,6 +430,7 @@ void rb_ary_store(VALUE array, long index, VALUE value);
 VALUE rb_ary_entry(VALUE array, long index);
 VALUE rb_ary_dup(VALUE array);
 VALUE rb_ary_each(VALUE array);
+VALUE rb_ary_unshift(VALUE array, VALUE value);
 #define rb_assoc_new(a, b) rb_ary_new3(2, a, b)
 VALUE rb_check_array_type(VALUE array);
 
@@ -420,7 +456,7 @@ VALUE rb_class_real(VALUE ruby_class);
 VALUE rb_class_superclass(VALUE ruby_class);
 VALUE rb_class_of(VALUE object);
 VALUE rb_obj_class(VALUE object);
-#define CLASS_OF(object) rb_class_of((VALUE) (object))
+VALUE CLASS_OF(VALUE object);
 VALUE rb_obj_alloc(VALUE ruby_class);
 VALUE rb_class_path(VALUE ruby_class);
 VALUE rb_path2class(const char *string);
@@ -434,6 +470,27 @@ VALUE rb_proc_new(void *function, VALUE value);
 void rb_warn(const char *fmt, ...);
 void rb_warning(const char *fmt, ...);
 
+MUST_INLINE int rb_jt_scan_args_0_HASH(int argc, VALUE *argv, const char *format, VALUE *v1) {
+  if (argc >= 1) *v1 = argv[0];
+  return argc;
+}
+
+MUST_INLINE int rb_jt_scan_args_02(int argc, VALUE *argv, const char *format, VALUE *v1, VALUE *v2) {
+  if (argc >= 1) *v1 = argv[0];
+  if (argc >= 2) *v2 = argv[1];
+  return argc;
+}
+
+MUST_INLINE int rb_jt_scan_args_11(int argc, VALUE *argv, const char *format, VALUE *v1, VALUE *v2) {
+  if (argc < 1) {
+    rb_jt_error("rb_jt_scan_args_11 error case not implemented");
+    abort();
+  }
+  *v1 = argv[0];
+  if (argc >= 2) *v2 = argv[1];
+  return argc - 1;
+}
+
 int rb_scan_args(int argc, VALUE *argv, const char *format, ...);
 
 // Calls
@@ -522,7 +579,7 @@ void rb_undef(VALUE module, ID name);
 
 void rb_attr(VALUE ruby_class, ID name, int read, int write, int ex);
 
-typedef VALUE (*rb_alloc_func_t)(VALUE);
+typedef VALUE (*rb_alloc_func_t)(VALUE ruby_class);
 void rb_define_alloc_func(VALUE ruby_class, rb_alloc_func_t alloc_function);
 
 // Mutexes
@@ -610,7 +667,7 @@ void rb_fd_fix_cloexec(int fd);
 
 int rb_jt_io_handle(VALUE file);
 
-#define GetOpenFile(file, pointer) ((pointer)->fd = rb_jt_io_handle(file))
+#define GetOpenFile(file, pointer) ((pointer) = truffle_managed_malloc(sizeof(rb_io_t)), (pointer)->fd = rb_jt_io_handle(file))
 
 int rb_io_wait_readable(int fd);
 int rb_io_wait_writable(int fd);
@@ -618,20 +675,26 @@ void rb_thread_wait_fd(int fd);
 
 NORETURN(void rb_eof_error(void));
 
+// Objects
+
+struct RBasic {
+  // Empty
+};
+
 // Data
 
 struct RData {
-  // No RBasic object header
+  struct RBasic basic;
   void (*dmark)(void *data);
   void (*dfree)(void *data);
   void *data;
 };
 
-struct RData *rb_jt_wrap_rdata(VALUE value);
+struct RData *rb_jt_adapt_rdata(VALUE value);
 
-#define RDATA(value) rb_jt_wrap_rdata(value)
+#define RDATA(value) rb_jt_adapt_rdata(value)
 
-#define DATA_PTR(value) *((volatile intptr_t*) 0)
+#define DATA_PTR(value) (RDATA(value)->data)
 
 // Typed data
 
@@ -650,8 +713,19 @@ struct rb_data_type_struct {
   VALUE flags;
 };
 
+struct RTypedData {
+  struct RBasic basic;
+  const rb_data_type_t *type;
+  VALUE typed_flag;
+  void *data;
+};
+
 #define RUBY_TYPED_FREE_IMMEDIATELY 1
 
+#define RTYPEDDATA(value) ((struct RTypedData *)RDATA(value))
+
+#define RTYPEDDATA_DATA(value) (RTYPEDDATA(value)->data)
+
 VALUE rb_data_typed_object_wrap(VALUE ruby_class, void *data, const rb_data_type_t *data_type);
 
 #define TypedData_Wrap_Struct(ruby_class, data_type, data) rb_data_typed_object_wrap((ruby_class), (data), (data_type))
@@ -670,8 +744,6 @@ void *rb_check_typeddata(VALUE value, const rb_data_type_t *data_type);
 
 #define TypedData_Get_Struct(value, type, data_type, variable) ((variable) = (type *)rb_check_typeddata((value), (data_type)))
 
-#define RTYPEDDATA_DATA(value) *((volatile int*) 0)
-
 // VM
 
 VALUE *rb_ruby_verbose_ptr(void);
