@@ -65,7 +65,7 @@ module Truffle
         cmd = Array(cmd) # wrap if it's just a string
 
         formatted_env, command = if cmd[0].is_a?(Hash)
-                                   [cmd[0].map { |k, v| "#{k}=#{v}" }, cmd[1..-1]]
+                                   [cmd[0].map { |k, v| "#{k}=\"#{v}\"" }, cmd[1..-1]]
                                  else
                                    [[], cmd]
                                  end
@@ -186,6 +186,8 @@ module Truffle
     ROOT              = Pathname(__FILE__).dirname.parent.parent.expand_path
     JRUBY_PATH        = ROOT.join('../../../..').expand_path
     JRUBY_BIN         = JRUBY_PATH.join('bin', 'jruby')
+    JT                = JRUBY_PATH.join('tool', 'jt.rb')
+    INTERPRETER_PATH  = JT.exist? ? [JT.to_s, "ruby"] : JRUBY_BIN.to_s
 
     module OptionBlocks
       STORE_NEW_VALUE         = -> (new, old, _) { new }
@@ -198,7 +200,7 @@ module Truffle
 
     begin
       apply_pattern = -> (pattern, old, options) do
-        Dir.glob(pattern) do |file|
+        Dir.glob(pattern).sort.each do |file|
           if options[:exclude_pattern].any? { |p| /#{p}/ =~ file }
             puts "skipped: #{file}"
             next
@@ -248,8 +250,7 @@ module Truffle
                   }.merge(shared_offline_options),
           run:    {
               help:             ['-h', '--help', 'Show this message', STORE_NEW_VALUE, false],
-              interpreter_path: ['--interpreter-path PATH', "Path to #{BRANDING} interpreter executable", STORE_NEW_VALUE,
-                                 JRUBY_BIN],
+              interpreter_path: ['--interpreter-path PATH', "Path to #{BRANDING} interpreter executable", STORE_NEW_VALUE, INTERPRETER_PATH],
               no_truffle:       ['-n', '--no-truffle', "Use conventional JRuby instead of #{BRANDING}", STORE_NEW_NEGATED_VALUE, false],
               graal:            ['-g', '--graal', 'Run on graal', STORE_NEW_VALUE, false],
               build:            ['-b', '--build', 'Run `jt build` using conventional JRuby', STORE_NEW_VALUE, false],
@@ -370,7 +371,7 @@ module Truffle
         @gem_name = if @options[:global][:configuration]
                       @options[:global][:configuration].to_sym
                     else
-                      candidates = Dir['*.gemspec']
+                      candidates = Dir['*.gemspec'].sort
                       if candidates.size == 1
                         gem_name, _ = candidates.first.split('.')
                         gem_name.to_sym
@@ -417,6 +418,11 @@ module Truffle
       @options[:global][:verbose]
     end
 
+    def interpreter_path
+      # Might be an Array or just a String
+      [*@options[:run][:interpreter_path]]
+    end
+
     def build_option_parsers
       option_parsers = OPTION_DEFINITIONS.each_with_object({}) do |(name, parser_options), parsers|
         parsers[name] = build_option_parser(parser_options, @options.fetch(name))
@@ -518,7 +524,7 @@ module Truffle
           gem_name, options = parse_gemfile_line(line)
           repo_name         = options[:git].split('/').last
           repo_match        = "#{gems_path}/bundler/gems/#{repo_name}-*"
-          repo_path         = Dir[repo_match].first
+          repo_path         = Dir[repo_match].sort.first
 
           ["# Overridden by jtr\n",
            '# ' + line,
@@ -575,6 +581,7 @@ module Truffle
       gemfile_use_path!(target_gem_path) if @options[:setup][:offline]
 
       execute_cmd([JRUBY_BIN.to_s,
+                   '-X-C', # See https://github.com/jruby/jruby/issues/4171
                    "#{Gem.bindir}/bundle",
                    *bundle_options,
                    'install',
@@ -603,7 +610,7 @@ module Truffle
     end
 
     def subcommand_run(rest)
-      jruby_path = Pathname("#{@options[:run][:interpreter_path]}/../..").expand_path
+      jruby_path = Pathname("#{interpreter_path[0]}/../..").expand_path
       raise unless jruby_path.absolute?
       ruby_options, rest = if rest.include?('--')
                              split = rest.index('--')
@@ -623,7 +630,7 @@ module Truffle
       end
 
       executable = if @options[:run][:executable]
-                     executables = Dir.glob("#{@options[:global][:truffle_bundle_path]}/jruby+truffle/*/gems/*/{bin,exe}/*")
+                     executables = Dir.glob("#{@options[:global][:truffle_bundle_path]}/jruby+truffle/*/gems/*/{bin,exe}/*").sort
                      executables.find { |path| File.basename(path) == @options[:run][:executable] } or
                          raise "no executable with name '#{@options[:run][:executable]}' found"
                    end
@@ -637,27 +644,31 @@ module Truffle
           '-X+T',
           "-J-Xmx#{@options[:run][:xmx]}",
           *(%w[-J-ea -J-esa] unless @options[:run][:no_asserts]),
-          ("-Xtruffle.core.load_path=#{core_load_path}" if @options[:global][:use_fs_core] && !missing_core_load_path),
-          ('-Xtruffle.exceptions.print_java=true' if @options[:run][:jexception])
+          *("-Xtruffle.core.load_path=#{core_load_path}" if @options[:global][:use_fs_core] && !missing_core_load_path),
+          *('-Xtruffle.exceptions.print_java=true' if @options[:run][:jexception])
       ]
 
-      bundler_setup = "./#{@options[:global][:truffle_bundle_path]}/bundler/setup.rb"
-      cmd_options   = [
-          *(truffle_options unless @options[:run][:no_truffle]),
-          (format(@options[:global][:debug_option], @options[:global][:debug_port]) if @options[:run][:debug]),
+      jruby_options = [
+          *(format(@options[:global][:debug_option], @options[:global][:debug_port]) if @options[:run][:debug]),
+          *(truffle_options unless @options[:run][:no_truffle])
+      ]
+
+      bundler_setup = File.expand_path "#{@options[:global][:truffle_bundle_path]}/bundler/setup.rb"
+      env_options   = [
           *ruby_options,
           *(['-r', bundler_setup] if File.exist? bundler_setup),
           *@options[:run][:load_path].flat_map { |v| ['-I', v] },
           *@options[:run][:require].flat_map { |v| ['-r', v] }
-      ].compact
+      ]
 
       env            = @options[:run][:environment]
       env['JAVACMD'] = @options[:global][:graal_path] if @options[:run][:graal]
       env.each { |k, v| env[k] = v.to_s }
 
       cmd = [(env unless env.empty?),
-             @options[:run][:interpreter_path].to_s,
-             *cmd_options,
+             *interpreter_path,
+             *jruby_options,
+             *env_options,
              executable,
              *rest
       ].compact
@@ -688,7 +699,7 @@ module Truffle
                             File.read(Pathname(@called_from_dir).join(path))
                           end
                         end
-        cis_to_run     = YAML.load(batch_content)
+        cis_to_run    = YAML.load(batch_content)
 
         results = cis_to_run.map do |ci|
           # ci is just a name or a array of name and options
